"""
Some utility functions.

Miscellaneous utilities

* list2set
* first
* uniq
* more_than

Term characterisation and generation

* to_term
* from_n3

Date/time utilities

* date_time
* parse_date_time

"""

from __future__ import annotations

from calendar import timegm
from os.path import splitext

# from time import daylight
from time import altzone, gmtime, localtime, time, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)
from urllib.parse import quote, urlsplit, urlunsplit

import rdflib.graph  # avoid circular dependency
import rdflib.namespace
import rdflib.term
from rdflib.compat import sign

if TYPE_CHECKING:
    from rdflib.graph import Graph


__all__ = [
    "list2set",
    "first",
    "uniq",
    "more_than",
    "to_term",
    "from_n3",
    "date_time",
    "parse_date_time",
    "guess_format",
    "find_roots",
    "get_tree",
    "_coalesce",
    "_iri2uri",
]

_HashableT = TypeVar("_HashableT", bound=Hashable)
_AnyT = TypeVar("_AnyT")


SUFFIX_FORMAT_MAP = {
    "xml": "xml",
    "rdf": "xml",
    "owl": "xml",
    "n3": "n3",
    "ttl": "turtle",
    "nt": "nt",
    "trix": "trix",
    "xhtml": "rdfa",
    "html": "rdfa",
    "svg": "rdfa",
    "nq": "nquads",
    "nquads": "nquads",
    "trig": "trig",
    "json": "json-ld",
    "jsonld": "json-ld",
    "json-ld": "json-ld",
}


FORMAT_MIMETYPE_MAP = {
    "xml": ["application/rdf+xml"],
    "n3": ["text/n3"],
    "turtle": ["text/turtle"],
    "nt": ["application/n-triples"],
    "trix": ["application/trix"],
    "rdfa": ["text/html", "application/xhtml+xml"],
    "nquads": ["application/n-quads"],
    "trig": ["application/trig"],
    "json-ld": ["application/ld+json"],
}


RESPONSE_TABLE_FORMAT_MIMETYPE_MAP = {
    "xml": ["application/sparql-results+xml"],
    "json": ["application/sparql-results+json"],
    "csv": ["text/csv"],
    "tsv": ["text/tab-separated-values"],
}


def list2set(seq: Iterable[_HashableT]) -> List[_HashableT]:
    """
    Return a new list without duplicates.
    Preserves the order, unlike set(seq)
    """
    seen = set()
    # type error: "add" of "set" does not return a value
    return [x for x in seq if x not in seen and not seen.add(x)]  # type: ignore[func-returns-value]


def first(seq: Iterable[_AnyT]) -> Optional[_AnyT]:
    """
    return the first element in a python sequence
    for graphs, use graph.value instead
    """
    for result in seq:
        return result
    return None


def uniq(sequence: Iterable[str], strip: int = 0) -> Set[str]:
    """removes duplicate strings from the sequence."""
    if strip:
        return set(s.strip() for s in sequence)
    else:
        return set(sequence)


def more_than(sequence: Iterable[Any], number: int) -> int:
    "Returns 1 if sequence has more items than number and 0 if not."
    i = 0
    for item in sequence:
        i += 1
        if i > number:
            return 1
    return 0


def to_term(
    s: Optional[str], default: Optional[rdflib.term.Identifier] = None
) -> Optional[rdflib.term.Identifier]:
    """
    Creates and returns an Identifier of type corresponding
    to the pattern of the given positional argument string `s`:

    '' returns the `default` keyword argument value or `None`

    '<s>' returns `URIRef(s)` (i.e. without angle brackets)

    '"s"' returns `Literal(s)` (i.e. without doublequotes)

    '_s' returns `BNode(s)` (i.e. without leading underscore)

    """
    if not s:
        return default
    elif s.startswith("<") and s.endswith(">"):
        return rdflib.term.URIRef(s[1:-1])
    elif s.startswith('"') and s.endswith('"'):
        return rdflib.term.Literal(s[1:-1])
    elif s.startswith("_"):
        return rdflib.term.BNode(s)
    else:
        msg = "Unrecognised term syntax: '%s'" % s
        raise Exception(msg)


def from_n3(
    s: str,
    default: Optional[str] = None,
    backend: Optional[str] = None,
    nsm: Optional[rdflib.namespace.NamespaceManager] = None,
) -> Optional[Union[rdflib.term.Node, str]]:
    r'''Creates the Identifier corresponding to the given n3 string.

    ```python
    >>> from rdflib.term import URIRef, Literal
    >>> from rdflib.namespace import NamespaceManager
    >>> from_n3('<http://ex.com/foo>') == URIRef('http://ex.com/foo')
    True
    >>> from_n3('"foo"@de') == Literal('foo', lang='de')
    True
    >>> from_n3('"""multi\nline\nstring"""@en') == Literal(
    ...     'multi\nline\nstring', lang='en')
    True
    >>> from_n3('42') == Literal(42)
    True
    >>> from_n3(Literal(42).n3()) == Literal(42)
    True
    >>> from_n3('"42"^^xsd:integer') == Literal(42)
    True
    >>> from rdflib import RDFS
    >>> from_n3('rdfs:label') == RDFS['label']
    True
    >>> nsm = NamespaceManager(rdflib.graph.Graph())
    >>> nsm.bind('dbpedia', 'http://dbpedia.org/resource/')
    >>> berlin = URIRef('http://dbpedia.org/resource/Berlin')
    >>> from_n3('dbpedia:Berlin', nsm=nsm) == berlin
    True

    ```
    '''
    if not s:
        return default
    if s.startswith("<"):
        # Hack: this should correctly handle strings with either native unicode
        # characters, or \u1234 unicode escapes.
        return rdflib.term.URIRef(
            s[1:-1].encode("raw-unicode-escape").decode("unicode-escape")
        )
    elif s.startswith('"'):
        if s.startswith('"""'):
            quotes = '"""'
        else:
            quotes = '"'
        value, rest = s.rsplit(quotes, 1)
        value = value[len(quotes) :]  # strip leading quotes
        datatype = None
        language = None

        # as a given datatype overrules lang-tag check for it first
        dtoffset = rest.rfind("^^")
        if dtoffset >= 0:
            # found a datatype
            # datatype has to come after lang-tag so ignore everything before
            # see: http://www.w3.org/TR/2011/WD-turtle-20110809/
            # #prod-turtle2-RDFLiteral
            datatype = from_n3(rest[dtoffset + 2 :], default, backend, nsm)
        else:
            if rest.startswith("@"):
                language = rest[1:]  # strip leading at sign

        value = value.replace(r"\"", '"')
        # unicode-escape interprets \xhh as an escape sequence,
        # but n3 does not define it as such.
        value = value.replace(r"\x", r"\\x")
        # Hack: this should correctly handle strings with either native unicode
        # characters, or \u1234 unicode escapes.
        value = value.encode("raw-unicode-escape").decode("unicode-escape")
        # type error: Argument 3 to "Literal" has incompatible type "Union[Node, str, None]"; expected "Optional[str]"
        return rdflib.term.Literal(value, language, datatype)  # type: ignore[arg-type]
    elif s == "true" or s == "false":
        return rdflib.term.Literal(s == "true")
    elif (
        s.lower()
        .replace(".", "", 1)
        .replace("-", "", 1)
        .replace("e", "", 1)
        .isnumeric()
    ):
        if "e" in s.lower():
            return rdflib.term.Literal(s, datatype=rdflib.namespace.XSD.double)
        if "." in s:
            return rdflib.term.Literal(float(s), datatype=rdflib.namespace.XSD.decimal)
        return rdflib.term.Literal(int(s), datatype=rdflib.namespace.XSD.integer)

    elif s.startswith("{"):
        identifier = from_n3(s[1:-1])
        # type error: Argument 1 to "QuotedGraph" has incompatible type "Optional[str]"; expected "Union[Store, str]"
        # type error: Argument 2 to "QuotedGraph" has incompatible type "Union[Node, str, None]"; expected "Union[IdentifiedNode, str, None]"
        return rdflib.graph.QuotedGraph(backend, identifier)  # type: ignore[arg-type]
    elif s.startswith("["):
        identifier = from_n3(s[1:-1])
        # type error: Argument 1 to "Graph" has incompatible type "Optional[str]"; expected "Union[Store, str]"
        # type error: Argument 2 to "Graph" has incompatible type "Union[Node, str, None]"; expected "Union[IdentifiedNode, str, None]"
        return rdflib.graph.Graph(backend, identifier)  # type: ignore[arg-type]
    elif s.startswith("_:"):
        return rdflib.term.BNode(s[2:])
    elif ":" in s:
        if nsm is None:
            # instantiate default NamespaceManager and rely on its defaults
            nsm = rdflib.namespace.NamespaceManager(rdflib.graph.Graph())
        prefix, last_part = s.split(":", 1)
        ns = dict(nsm.namespaces())[prefix]
        return rdflib.namespace.Namespace(ns)[last_part]
    else:
        return rdflib.term.BNode(s)


def date_time(t=None, local_time_zone=False):
    """http://www.w3.org/TR/NOTE-datetime ex: 1997-07-16T19:20:30Z

    ```python
    >>> date_time(1126482850)
    '2005-09-11T23:54:10Z'

    @@ this will change depending on where it is run
    #>>> date_time(1126482850, local_time_zone=True)
    #'2005-09-11T19:54:10-04:00'

    >>> date_time(1)
    '1970-01-01T00:00:01Z'

    >>> date_time(0)
    '1970-01-01T00:00:00Z'

    ```
    """
    if t is None:
        t = time()

    if local_time_zone:
        time_tuple = localtime(t)
        if time_tuple[8]:
            tz_mins = altzone // 60
        else:
            tz_mins = timezone // 60
        tzd = "-%02d:%02d" % (tz_mins // 60, tz_mins % 60)
    else:
        time_tuple = gmtime(t)
        tzd = "Z"

    year, month, day, hh, mm, ss, wd, y, z = time_tuple
    s = "%0004d-%02d-%02dT%02d:%02d:%02d%s" % (year, month, day, hh, mm, ss, tzd)
    return s


def parse_date_time(val: str) -> int:
    """always returns seconds in UTC

    ```python
    # tests are written like this to make any errors easier to understand
    >>> parse_date_time('2005-09-11T23:54:10Z') - 1126482850.0
    0.0

    >>> parse_date_time('2005-09-11T16:54:10-07:00') - 1126482850.0
    0.0

    >>> parse_date_time('1970-01-01T00:00:01Z') - 1.0
    0.0

    >>> parse_date_time('1970-01-01T00:00:00Z') - 0.0
    0.0
    >>> parse_date_time("2005-09-05T10:42:00") - 1125916920.0
    0.0

    ```
    """

    if "T" not in val:
        val += "T00:00:00Z"

    ymd, time = val.split("T")
    hms, tz_str = time[0:8], time[8:]

    if not tz_str or tz_str == "Z":
        time = time[:-1]
        tz_offset = 0
    else:
        signed_hrs = int(tz_str[:3])
        mins = int(tz_str[4:6])
        secs = (sign(signed_hrs) * mins + signed_hrs * 60) * 60
        tz_offset = -secs

    year, month, day = ymd.split("-")
    hour, minute, second = hms.split(":")

    t = timegm(
        (int(year), int(month), int(day), int(hour), int(minute), int(second), 0, 0, 0)
    )
    t = t + tz_offset
    return t


def guess_format(fpath: str, fmap: Optional[Dict[str, str]] = None) -> Optional[str]:
    """
    Guess RDF serialization based on file suffix. Uses
    `SUFFIX_FORMAT_MAP` unless `fmap` is provided.

    Example:
        ```python
        >>> guess_format('path/to/file.rdf')
        'xml'
        >>> guess_format('path/to/file.owl')
        'xml'
        >>> guess_format('path/to/file.ttl')
        'turtle'
        >>> guess_format('path/to/file.json')
        'json-ld'
        >>> guess_format('path/to/file.xhtml')
        'rdfa'
        >>> guess_format('path/to/file.svg')
        'rdfa'
        >>> guess_format('path/to/file.xhtml', {'xhtml': 'grddl'})
        'grddl'

        ```

        This also works with just the suffixes, with or without leading dot, and
        regardless of letter case:

        ```python
        >>> guess_format('.rdf')
        'xml'
        >>> guess_format('rdf')
        'xml'
        >>> guess_format('RDF')
        'xml'

        ```
    """
    fmap = fmap or SUFFIX_FORMAT_MAP
    return fmap.get(_get_ext(fpath)) or fmap.get(fpath.lower())


def _get_ext(fpath: str, lower: bool = True) -> str:
    """
    Gets the file extension from a file(path); stripped of leading '.' and in
    lower case.

    Example:
        ```python
        >>> _get_ext("path/to/file.txt")
        'txt'
        >>> _get_ext("OTHER.PDF")
        'pdf'
        >>> _get_ext("noext")
        ''
        >>> _get_ext(".rdf")
        'rdf'

        ```
    """
    ext = splitext(fpath)[-1]
    if ext == "" and fpath.startswith("."):
        ext = fpath
    if lower:
        ext = ext.lower()
    if ext.startswith("."):
        ext = ext[1:]
    return ext


def find_roots(
    graph: Graph,
    prop: rdflib.term.URIRef,
    roots: Optional[Set[rdflib.term.Node]] = None,
) -> Set[rdflib.term.Node]:
    """Find the roots in some sort of transitive hierarchy.

    find_roots(graph, rdflib.RDFS.subClassOf)
    will return a set of all roots of the sub-class hierarchy

    Assumes triple of the form (child, prop, parent), i.e. the direction of
    `RDFS.subClassOf` or `SKOS.broader`
    """

    non_roots: Set[rdflib.term.Node] = set()
    if roots is None:
        roots = set()
    for x, y in graph.subject_objects(prop):
        non_roots.add(x)
        if x in roots:
            roots.remove(x)
        if y not in non_roots:
            roots.add(y)
    return roots


def get_tree(
    graph: Graph,
    root: rdflib.term.Node,
    prop: rdflib.term.URIRef,
    mapper: Callable[[rdflib.term.Node], rdflib.term.Node] = lambda x: x,
    sortkey: Optional[Callable[[Any], Any]] = None,
    done: Optional[Set[rdflib.term.Node]] = None,
    dir: str = "down",
) -> Optional[Tuple[rdflib.term.Node, List[Any]]]:
    """
    Return a nested list/tuple structure representing the tree
    built by the transitive property given, starting from the root given

    i.e.

    ```python
    get_tree(
        graph,
        rdflib.URIRef("http://xmlns.com/foaf/0.1/Person"),
        rdflib.RDFS.subClassOf,
    )
    ```

    will return the structure for the subClassTree below person.

    dir='down' assumes triple of the form (child, prop, parent),
    i.e. the direction of RDFS.subClassOf or SKOS.broader
    Any other dir traverses in the other direction
    """

    if done is None:
        done = set()
    if root in done:
        # type error: Return value expected
        return  # type: ignore[return-value]
    done.add(root)
    tree = []

    branches: Iterator[rdflib.term.Node]
    if dir == "down":
        branches = graph.subjects(prop, root)
    else:
        branches = graph.objects(root, prop)

    for branch in branches:
        t = get_tree(graph, branch, prop, mapper, sortkey, done, dir)
        if t:
            tree.append(t)

    return (mapper(root), sorted(tree, key=sortkey))


@overload
def _coalesce(*args: Optional[_AnyT], default: _AnyT) -> _AnyT: ...


@overload
def _coalesce(
    *args: Optional[_AnyT], default: Optional[_AnyT] = ...
) -> Optional[_AnyT]: ...


def _coalesce(
    *args: Optional[_AnyT], default: Optional[_AnyT] = None
) -> Optional[_AnyT]:
    """
    This is a null coalescing function, it will return the first non-`None`
    argument passed to it, otherwise it will return `default` which is `None`
    by default.

    For more info regarding the rationale of this function see deferred
    [PEP 505](https://peps.python.org/pep-0505/).

    Args:
        *args: Values to consider as candidates to return, the first arg that
            is not `None` will be returned. If no argument is passed this function
            will return None.
        default: The default value to return if none of the args are not `None`.

    Returns:
        The first `args` that is not `None`, otherwise the value of
            `default` if there are no `args` or if all `args` are `None`.
    """
    for arg in args:
        if arg is not None:
            return arg
    return default


_RFC3986_SUBDELIMS = "!$&'()*+,;="
"""
`sub-delims` production from
[RFC 3986, section 2.2](https://www.rfc-editor.org/rfc/rfc3986.html#section-2.2).
"""

_RFC3986_PCHAR_NU = "%" + _RFC3986_SUBDELIMS + ":@"
"""
The non-unreserved characters in the `pchar` production from RFC 3986.
"""

_QUERY_SAFE_CHARS = _RFC3986_PCHAR_NU + "/?"
"""
The non-unreserved characters that are safe to use in in the query and fragment
components.

```
pchar         = unreserved / pct-encoded / sub-delims / ":" / "@" query
= *( pchar / "/" / "?" ) fragment      = *( pchar / "/" / "?" )
```
"""

_USERNAME_SAFE_CHARS = _RFC3986_SUBDELIMS + "%"
"""
The non-unreserved characters that are safe to use in the username and password
components.

```
userinfo      = *( unreserved / pct-encoded / sub-delims / ":" )
```

":" is excluded as this is only used for the username and password components,
and they are treated separately.
"""

_PATH_SAFE_CHARS = _RFC3986_PCHAR_NU + "/"
"""
The non-unreserved characters that are safe to use in the path component.

This is based on various path-related productions from RFC 3986.
"""


def _iri2uri(iri: str) -> str:
    """
    Prior art:

    - [iri_to_uri from Werkzeug](https://github.com/pallets/werkzeug/blob/92c6380248c7272ee668e1f8bbd80447027ccce2/src/werkzeug/urls.py#L926-L931)

    ```python
    >>> _iri2uri("https://dbpedia.org/resource/Almería")
    'https://dbpedia.org/resource/Almer%C3%ADa'

    ```
    """
    # https://datatracker.ietf.org/doc/html/rfc3986
    # https://datatracker.ietf.org/doc/html/rfc3305

    parts = urlsplit(iri)
    (scheme, netloc, path, query, fragment) = parts

    # Just support http/https, otherwise return the iri unaltered
    if scheme not in ["http", "https"]:
        return iri

    path = quote(path, safe=_PATH_SAFE_CHARS)
    query = quote(query, safe=_QUERY_SAFE_CHARS)
    fragment = quote(fragment, safe=_QUERY_SAFE_CHARS)

    if parts.hostname:
        netloc = parts.hostname.encode("idna").decode("ascii")
    else:
        netloc = ""

    if ":" in netloc:
        # Quote IPv6 addresses
        netloc = f"[{netloc}]"

    if parts.port:
        netloc = f"{netloc}:{parts.port}"

    if parts.username:
        auth = quote(parts.username, safe=_USERNAME_SAFE_CHARS)
        if parts.password:
            pass_quoted = quote(parts.password, safe=_USERNAME_SAFE_CHARS)
            auth = f"{auth}:{pass_quoted}"
        netloc = f"{auth}@{netloc}"

    uri = urlunsplit((scheme, netloc, path, query, fragment))

    if iri.endswith("#") and not uri.endswith("#"):
        uri += "#"

    return uri



from __future__ import annotations

from textwrap import dedent
from typing import Any, Generator, Iterable, Iterator, Mapping, Optional, Tuple

from rdflib import Graph
from rdflib.contrib.rdf4j import has_httpx
from rdflib.contrib.rdf4j.exceptions import RepositoryNotFoundError
from rdflib.graph import (
    DATASET_DEFAULT_GRAPH_ID,
    Dataset,
    _ContextType,
    _QuadType,
    _TriplePatternType,
    _TripleType,
)
from rdflib.store import VALID_STORE, Store
from rdflib.term import BNode, Node, URIRef, Variable

if has_httpx:
    from rdflib.contrib.rdf4j import RDF4JClient


def _inject_prefixes(query: str, extra_bindings: Mapping[str, Any]) -> str:
    bindings = set(list(extra_bindings.items()))
    if not bindings:
        return query
    return "\n".join(
        [
            "\n".join(["PREFIX %s: <%s>" % (k, v) for k, v in bindings]),
            "",  # separate ns_bindings from query with an empty line
            query,
        ]
    )


def _node_to_sparql(node: Node) -> str:
    if isinstance(node, BNode):
        raise Exception(
            "SPARQL-based stores do not support BNodes"
            "See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes"
        )
    return node.n3()


def _default_repo_config(repository_id: str) -> str:
    return dedent(
        f"""
        PREFIX config: <tag:rdf4j.org,2023:config/>

        []    a config:Repository ;
            config:rep.id "{repository_id}" ;
            config:rep.impl
                [
                    config:rep.type "openrdf:SailRepository" ;
                    config:sail.impl
                        [
                            config:native.tripleIndexers "spoc,posc" ;
                            config:sail.defaultQueryEvaluationMode "STANDARD" ;
                            config:sail.iterationCacheSyncThreshold "10000" ;
                            config:sail.type "openrdf:NativeStore" ;
                        ] ;
                ] ;
        .
    """
    )


class RDF4JStore(Store):
    """An RDF4J store."""

    context_aware = True
    formula_aware = False
    transaction_aware = False
    graph_aware = True

    def __init__(
        self,
        base_url: str,
        repository_id: str,
        configuration: str | None = None,
        auth: tuple[str, str] | None = None,
        timeout: float = 30.0,
        create: bool = False,
        **kwargs,
    ):
        if configuration is None:
            configuration = _default_repo_config(repository_id)
        self._client = RDF4JClient(base_url, auth, timeout, **kwargs)
        self._repository_id = repository_id
        self._repo = None
        self.open(configuration, create)
        super().__init__()

    @property
    def client(self):
        return self._client

    @property
    def repo(self):
        if self._repo is None:
            self._repo = self.client.repositories.get(self._repository_id)
        return self._repo

    def open(
        self, configuration: str | tuple[str, str] | None, create: bool = False
    ) -> int | None:
        try:
            # Try connecting to the repository.
            self.repo.health()
        except RepositoryNotFoundError:
            if create:
                self.client.repositories.create(self._repository_id, configuration)
                self.repo.health()
            else:
                raise Exception(f"Repository {self._repository_id} not found.")

        return VALID_STORE

    def close(self, commit_pending_transaction: bool = False) -> None:
        self.client.close()

    def add(
        self,
        triple: _TripleType,
        context: _ContextType | None = None,
        quoted: bool = False,
    ) -> None:
        s, p, o = triple
        graph_name = (
            ""
            if context is None or context.identifier == DATASET_DEFAULT_GRAPH_ID
            else context.identifier.n3()
        )
        statement = f"{s.n3()} {p.n3()} {o.n3()} {graph_name} ."
        self.repo.upload(statement)

    def addN(self, quads: Iterable[_QuadType]) -> None:  # noqa: N802
        statements = ""
        for s, p, o, c in quads:
            graph_name = (
                ""
                if c is None or c.identifier == DATASET_DEFAULT_GRAPH_ID
                else c.identifier.n3()
            )
            statement = f"{s.n3()} {p.n3()} {o.n3()} {graph_name} .\n"
            statements += statement
        self.repo.upload(statements)

    def remove(
        self,
        triple: _TriplePatternType,
        context: Optional[_ContextType] = None,
    ) -> None:
        s, p, o = triple
        g = context.identifier if context is not None else None
        self.repo.delete(s, p, o, g)

    def triples(
        self,
        triple_pattern: _TriplePatternType,
        context: Optional[_ContextType] = None,
    ) -> Iterator[Tuple[_TripleType, Iterator[Optional[_ContextType]]]]:
        s, p, o = triple_pattern
        graph_name = context.identifier if context is not None else None
        result_graph = self.repo.get(s, p, o, graph_name)
        if isinstance(result_graph, Dataset):
            for s, p, o, g in result_graph:
                yield (s, p, o), iter([Graph(self, identifier=g)])
        else:
            # It's a Graph object.
            for triple in result_graph:
                # Returning None for _ContextType as it's not used by the caller.
                yield triple, iter([None])

    def contexts(
        self, triple: Optional[_TripleType] = None
    ) -> Generator[_ContextType, None, None]:
        if triple is None:
            for graph_name in self.repo.graph_names():
                yield Graph(self, identifier=graph_name)
        else:
            s, p, o = triple
            params = (
                _node_to_sparql(s if s else Variable("s")),
                _node_to_sparql(p if p else Variable("p")),
                _node_to_sparql(o if o else Variable("o")),
            )
            query = (
                "SELECT DISTINCT ?graph WHERE { GRAPH ?graph { %s %s %s } }" % params
            )
            result = self.repo.query(query)
            for row in result:
                yield Graph(self, identifier=row["graph"])

    def bind(self, prefix: str, namespace: URIRef, override: bool = True) -> None:
        # Note: RDF4J namespaces always override.
        self.repo.namespaces.set(prefix, namespace)

    def prefix(self, namespace: URIRef) -> Optional[str]:
        namespace_prefixes = dict(
            [(x.namespace, x.prefix) for x in self.repo.namespaces.list()]
        )
        return namespace_prefixes.get(str(namespace))

    def namespace(self, prefix: str) -> Optional[URIRef]:
        result = self.repo.namespaces.get(prefix)
        return URIRef(result) if result is not None else None

    def namespaces(self) -> Iterator[Tuple[str, URIRef]]:
        for result in self.repo.namespaces.list():
            yield result.prefix, URIRef(result.namespace)

    def add_graph(self, graph: Graph) -> None:
        if graph.identifier != DATASET_DEFAULT_GRAPH_ID:
            # Note: this is a no-op since RDF4J doesn't support empty named graphs.
            self.repo.update(f"CREATE SILENT GRAPH {graph.identifier.n3()}")

    def remove_graph(self, graph: Graph) -> None:
        self.repo.graphs.clear(graph.identifier)

    def __len__(self, context: _ContextType | None = None) -> int:
        return self.repo.size(context if context is None else context.identifier)



"""RDF4J client module."""

from __future__ import annotations

import contextlib
import io
import typing as t
from dataclasses import dataclass
from typing import Any, BinaryIO, Iterable

import httpx

from rdflib import BNode
from rdflib.contrib.rdf4j.exceptions import (
    RDF4JUnsupportedProtocolError,
    RDFLibParserError,
    RepositoryAlreadyExistsError,
    RepositoryError,
    RepositoryNotFoundError,
    RepositoryNotHealthyError,
    RepositoryResponseFormatError,
    TransactionClosedError,
    TransactionCommitError,
    TransactionPingError,
    TransactionRollbackError,
)
from rdflib.contrib.rdf4j.util import (
    build_context_param,
    build_infer_param,
    build_sparql_query_accept_header,
    build_spo_param,
    rdf_payload_to_stream,
    validate_graph_name,
    validate_no_bnodes,
)
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, Dataset, Graph
from rdflib.query import Result
from rdflib.term import IdentifiedNode, Literal, URIRef

SubjectType = t.Union[URIRef, None]
PredicateType = t.Union[URIRef, None]
ObjectType = t.Union[URIRef, Literal, None]


@dataclass(frozen=True)
class NamespaceListingResult:
    """RDF4J namespace and prefix name result."""

    prefix: str
    namespace: str


class RDF4JNamespaceManager:
    """A namespace manager for RDF4J repositories.

    Parameters:
        identifier: The identifier of the repository.
        http_client: The httpx.Client instance.
    """

    def __init__(self, identifier: str, http_client: httpx.Client):
        self._identifier = identifier
        self._http_client = http_client

    @property
    def http_client(self):
        return self._http_client

    @property
    def identifier(self):
        """Repository identifier."""
        return self._identifier

    def list(self) -> list[NamespaceListingResult]:
        """List all namespace declarations in the repository.

        Returns:
            list[NamespaceListingResult]: List of namespace and prefix name results.

        Raises:
            RepositoryResponseFormatError: If the response format is unrecognized.
        """
        headers = {
            "Accept": "application/sparql-results+json",
        }
        response = self.http_client.get(
            f"/repositories/{self.identifier}/namespaces", headers=headers
        )
        response.raise_for_status()

        try:
            data = response.json()
            results = data["results"]["bindings"]
            return [
                NamespaceListingResult(
                    prefix=row["prefix"]["value"],
                    namespace=row["namespace"]["value"],
                )
                for row in results
            ]
        except (KeyError, ValueError) as err:
            raise RepositoryResponseFormatError(f"Unrecognised response format: {err}")

    def clear(self):
        """Clear all namespace declarations in the repository."""
        headers = {
            "Accept": "application/sparql-results+json",
        }
        response = self.http_client.delete(
            f"/repositories/{self.identifier}/namespaces", headers=headers
        )
        response.raise_for_status()

    def get(self, prefix: str) -> str | None:
        """Get the namespace URI for a given prefix.

        Parameters:
            prefix: The prefix to lookup.

        Returns:
            The namespace URI or `None` if not found.
        """
        if not prefix:
            raise ValueError("Prefix cannot be empty.")
        headers = {
            "Accept": "text/plain",
        }
        try:
            response = self.http_client.get(
                f"/repositories/{self.identifier}/namespaces/{prefix}", headers=headers
            )
            response.raise_for_status()
            return response.text
        except httpx.HTTPStatusError as err:
            if err.response.status_code == 404:
                return None
            raise

    def set(self, prefix: str, namespace: str):
        """Set the namespace URI for a given prefix.

        !!! note
            If the prefix was previously mapped to a different namespace, this will be
            overwritten.

        Parameters:
            prefix: The prefix to set.
            namespace: The namespace URI to set.
        """
        if not prefix:
            raise ValueError("Prefix cannot be empty.")
        if not namespace:
            raise ValueError("Namespace cannot be empty.")
        headers = {
            "Content-Type": "text/plain",
        }
        response = self.http_client.put(
            f"/repositories/{self.identifier}/namespaces/{prefix}",
            headers=headers,
            content=namespace,
        )
        response.raise_for_status()

    def remove(self, prefix: str):
        """Remove the namespace declaration for a given prefix.

        Parameters:
            prefix: The prefix to remove.
        """
        if not prefix:
            raise ValueError("Prefix cannot be empty.")
        response = self.http_client.delete(
            f"/repositories/{self.identifier}/namespaces/{prefix}"
        )
        response.raise_for_status()


class GraphStoreManager:
    """An RDF4J Graph Store Protocol Client.

    Parameters:
        identifier: The identifier of the repository.
        http_client: The httpx.Client instance.
    """

    def __init__(self, identifier: str, http_client: httpx.Client):
        self._identifier = identifier
        self._http_client = http_client
        self._content_type = "application/n-triples"

    @property
    def http_client(self):
        return self._http_client

    @property
    def identifier(self):
        """Repository identifier."""
        return self._identifier

    @staticmethod
    def _build_graph_name_params(graph_name: URIRef | str):
        params = {}
        if (
            isinstance(graph_name, URIRef)
            and graph_name == DATASET_DEFAULT_GRAPH_ID
            or isinstance(graph_name, str)
            and graph_name == str(DATASET_DEFAULT_GRAPH_ID)
        ):
            # Do nothing; GraphDB does not work with `?default=`
            # (note the trailing equal character), which is the default
            # behavior of httpx when setting the param value to an empty string.
            # httpx completely omits query parameters whose values are `None`, so that's
            # not an option either.
            # The workaround is to construct our own query parameter URL when we target
            # the default graph.
            pass
        else:
            params["graph"] = str(graph_name)
        return params

    def _build_url(self, graph_name: URIRef | str):
        url = f"/repositories/{self.identifier}/rdf-graphs/service"
        if isinstance(graph_name, URIRef) and graph_name == DATASET_DEFAULT_GRAPH_ID:
            url += "?default"
        return url

    def get(self, graph_name: URIRef | str) -> Graph:
        """Fetch all statements in the specified graph.

        Parameters:
            graph_name: The graph name of the graph.

                For the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

        Returns:
            A [`Graph`][rdflib.graph.Graph] object containing all statements in the
                graph.
        """
        if not graph_name:
            raise ValueError("Graph name must be provided.")
        validate_graph_name(graph_name)
        headers = {
            "Accept": self._content_type,
        }
        params = self._build_graph_name_params(graph_name) or None

        response = self.http_client.get(
            self._build_url(graph_name),
            headers=headers,
            params=params,
        )
        response.raise_for_status()

        return Graph(identifier=graph_name).parse(
            data=response.text, format=self._content_type
        )

    def add(self, graph_name: URIRef | str, data: str | bytes | BinaryIO | Graph):
        """Add statements to the specified graph.

        Parameters:
            graph_name: The graph name of the graph.

                For the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

            data: The RDF data to add.
        """
        if not graph_name:
            raise ValueError("Graph name must be provided.")
        validate_graph_name(graph_name)
        stream, should_close = rdf_payload_to_stream(data)
        headers = {
            "Content-Type": self._content_type,
        }
        params = self._build_graph_name_params(graph_name) or None
        try:
            response = self.http_client.post(
                self._build_url(graph_name),
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()

    def overwrite(self, graph_name: URIRef | str, data: str | bytes | BinaryIO | Graph):
        """Overwrite statements in the specified graph.

        Parameters:
            graph_name: The graph name of the graph.

                For the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

            data: The RDF data to overwrite with.
        """
        if not graph_name:
            raise ValueError("Graph name must be provided.")
        validate_graph_name(graph_name)
        stream, should_close = rdf_payload_to_stream(data)
        headers = {
            "Content-Type": self._content_type,
        }
        params = self._build_graph_name_params(graph_name) or None
        try:
            response = self.http_client.put(
                self._build_url(graph_name),
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()

    def clear(self, graph_name: URIRef | str):
        """Clear all statements in the specified graph.

        Parameters:
            graph_name: The graph name of the graph.

                For the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].
        """
        if not graph_name:
            raise ValueError("Graph name must be provided.")
        validate_graph_name(graph_name)
        params = self._build_graph_name_params(graph_name) or None
        response = self.http_client.delete(self._build_url(graph_name), params=params)
        response.raise_for_status()


@dataclass(frozen=True)
class RepositoryListingResult:
    """RDF4J repository listing result.

    Parameters:
        identifier: Repository identifier.
        uri: Repository URI.
        readable: Whether the repository is readable by the client.
        writable: Whether the repository is writable by the client.
        title: Repository title.
    """

    identifier: str
    uri: str
    readable: bool
    writable: bool
    title: str | None = None


class Repository:
    """RDF4J repository client.

    Parameters:
        identifier: The identifier of the repository.
        http_client: The httpx.Client instance.
    """

    def __init__(self, identifier: str, http_client: httpx.Client):
        self._identifier = identifier
        self._http_client = http_client
        self._namespace_manager: RDF4JNamespaceManager | None = None
        self._graph_store_manager: GraphStoreManager | None = None

    @property
    def http_client(self):
        return self._http_client

    @property
    def identifier(self):
        """Repository identifier."""
        return self._identifier

    @property
    def namespaces(self) -> RDF4JNamespaceManager:
        """Namespace manager for the repository."""
        if self._namespace_manager is None:
            self._namespace_manager = RDF4JNamespaceManager(
                self.identifier, self.http_client
            )
        return self._namespace_manager

    @property
    def graphs(self) -> GraphStoreManager:
        """Graph store manager for the repository."""
        if self._graph_store_manager is None:
            self._graph_store_manager = GraphStoreManager(
                self.identifier, self.http_client
            )
        return self._graph_store_manager

    def health(self) -> bool:
        """Repository health check.

        Returns:
            bool: True if the repository is healthy, otherwise an error is raised.

        Raises:
            RepositoryNotFoundError: If the repository is not found.
            RepositoryNotHealthyError: If the repository is not healthy.
        """
        headers = {
            "Content-Type": "application/sparql-query",
            "Accept": "application/sparql-results+json",
        }
        try:
            response = self.http_client.post(
                f"/repositories/{self._identifier}", headers=headers, content="ASK {}"
            )
            response.raise_for_status()
            return True
        except httpx.HTTPStatusError as err:
            if err.response.status_code == 404:
                raise RepositoryNotFoundError(
                    f"Repository {self._identifier} not found."
                )
            raise RepositoryNotHealthyError(
                f"Repository {self._identifier} is not healthy. {err.response.status_code} - {err.response.text}"
            )

    def size(self, graph_name: URIRef | Iterable[URIRef] | str | None = None) -> int:
        """The number of statements in the repository or in the specified graph name.

        Parameters:
            graph_name: Graph name(s) to restrict to.

                The default value `None` queries all graphs.

                To query just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

        Returns:
            The number of statements.

        Raises:
            RepositoryResponseFormatError: Fails to parse the repository size.
        """
        validate_graph_name(graph_name)
        params: dict[str, str] = {}
        build_context_param(params, graph_name)
        response = self.http_client.get(
            f"/repositories/{self.identifier}/size", params=params
        )
        response.raise_for_status()
        return self._to_size(response.text)

    @staticmethod
    def _to_size(size: str):
        try:
            value = int(size)
            if value >= 0:
                return value
            raise ValueError(f"Invalid repository size: {value}")
        except ValueError as err:
            raise RepositoryResponseFormatError(
                f"Failed to parse repository size: {err}"
            ) from err

    def query(self, query: str, **kwargs):
        """Execute a SPARQL query against the repository.

        !!! note
            A POST request is used by default. If any keyword arguments are provided,
            a GET request is used instead, and the arguments are passed as query parameters.

        Parameters:
            query: The SPARQL query to execute.
            **kwargs: Additional keyword arguments to include as query parameters
                in the request. See
                [RDF4J REST API - Execute SPARQL query](https://rdf4j.org/documentation/reference/rest-api/#tag/SPARQL/paths/~1repositories~1%7BrepositoryID%7D/get)
                for the list of supported query parameters.
        """
        headers = {"Content-Type": "application/sparql-query"}
        build_sparql_query_accept_header(query, headers)

        if not kwargs:
            response = self.http_client.post(
                f"/repositories/{self.identifier}", headers=headers, content=query
            )
        else:
            response = self.http_client.get(
                f"/repositories/{self.identifier}",
                headers=headers,
                params={"query": query, **kwargs},
            )
        response.raise_for_status()
        try:
            return Result.parse(
                io.BytesIO(response.content),
                content_type=response.headers["Content-Type"].split(";")[0],
            )
        except KeyError as err:
            raise RDFLibParserError(
                f"Failed to parse SPARQL query result {response.headers.get('Content-Type')}: {err}"
            ) from err

    def update(self, query: str):
        """Execute a SPARQL update operation on the repository.

        Parameters:
            query: The SPARQL update query to execute.
        """
        headers = {"Content-Type": "application/sparql-update"}
        response = self.http_client.post(
            f"/repositories/{self.identifier}/statements",
            headers=headers,
            content=query,
        )
        response.raise_for_status()

    def graph_names(self) -> list[IdentifiedNode]:
        """Get a list of all graph names in the repository.

        Returns:
            A list of graph names.

        Raises:
            RepositoryResponseFormatError: Fails to parse the repository graph names.
        """
        headers = {
            "Accept": "application/sparql-results+json",
        }
        response = self.http_client.get(
            f"/repositories/{self.identifier}/contexts", headers=headers
        )
        response.raise_for_status()
        try:
            values: list[IdentifiedNode] = []
            for row in response.json()["results"]["bindings"]:
                value = row["contextID"]["value"]
                value_type = row["contextID"]["type"]
                if value_type == "uri":
                    values.append(URIRef(value))
                elif value_type == "bnode":
                    values.append(BNode(value))
                else:
                    raise ValueError(f"Invalid graph name type: {value_type}")
            return values
        except Exception as err:
            raise RepositoryResponseFormatError(
                f"Failed to parse repository graph names: {err}"
            ) from err

    def get(
        self,
        subj: SubjectType = None,
        pred: PredicateType = None,
        obj: ObjectType = None,
        graph_name: URIRef | Iterable[URIRef] | str | None = None,
        infer: bool = True,
        content_type: str | None = None,
    ) -> Graph | Dataset:
        """Get RDF statements from the repository matching the filtering parameters.

        !!! Note
            The terms for `subj`, `pred`, `obj` or `graph_name` cannot be
            [`BNodes`][rdflib.term.BNode].

        Parameters:
            subj: Subject of the statement to filter by, or `None` to match all.
            pred: Predicate of the statement to filter by, or `None` to match all.
            obj: Object of the statement to filter by, or `None` to match all.
            graph_name: Graph name(s) to restrict to.

                The default value `None` queries all graphs.

                To query just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

            infer: Specifies whether inferred statements should be included in the
                result.
            content_type: The content type of the response.
                A triple-based format returns a [Graph][rdflib.graph.Graph], while a
                quad-based format returns a [`Dataset`][rdflib.graph.Dataset].

        Returns:
            A [`Graph`][rdflib.graph.Graph] or [`Dataset`][rdflib.graph.Dataset] object
                with the repository namespace prefixes bound to it.
        """
        validate_no_bnodes(subj, pred, obj, graph_name)
        if content_type is None:
            content_type = "application/n-quads"
        headers = {"Accept": content_type}
        params: dict[str, str] = {}
        build_context_param(params, graph_name)
        build_spo_param(params, subj, pred, obj)
        build_infer_param(params, infer=infer)

        response = self.http_client.get(
            f"/repositories/{self.identifier}/statements",
            headers=headers,
            params=params,
        )
        response.raise_for_status()
        triple_formats = [
            "application/n-triples",
            "text/turtle",
            "application/rdf+xml",
        ]
        try:
            if content_type in triple_formats:
                retval = Graph().parse(data=response.text, format=content_type)
            else:
                retval = Dataset().parse(data=response.text, format=content_type)
            for result in self.namespaces.list():
                retval.bind(result.prefix, result.namespace, replace=True)
            return retval
        except Exception as err:
            raise RDFLibParserError(f"Error parsing RDF: {err}") from err

    def upload(
        self,
        data: str | bytes | BinaryIO | Graph | Dataset,
        base_uri: str | None = None,
        content_type: str | None = None,
    ):
        """Upload and append statements to the repository.

        Parameters:
            data: The RDF data to upload.
            base_uri: The base URI to resolve against for any relative URIs in the data.
            content_type: The content type of the data. Defaults to
                `application/n-quads` when the value is `None`.
        """
        stream, should_close = rdf_payload_to_stream(data)
        try:
            headers = {"Content-Type": content_type or "application/n-quads"}
            params = {}
            if base_uri is not None:
                params["baseURI"] = base_uri
            response = self.http_client.post(
                f"/repositories/{self.identifier}/statements",
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()

    def overwrite(
        self,
        data: str | bytes | BinaryIO | Graph | Dataset,
        graph_name: URIRef | Iterable[URIRef] | str | None = None,
        base_uri: str | None = None,
        content_type: str | None = None,
    ):
        """Upload and overwrite statements in the repository.

        Parameters:
            data: The RDF data to upload.
            graph_name: Graph name(s) to restrict to.

                The default value `None` applies to all graphs.

                To apply to just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

            base_uri: The base URI to resolve against for any relative URIs in the data.
            content_type: The content type of the data. Defaults to
                `application/n-quads` when the value is `None`.
        """
        stream, should_close = rdf_payload_to_stream(data)
        validate_graph_name(graph_name)
        try:
            headers = {"Content-Type": content_type or "application/n-quads"}
            params: dict[str, str] = {}
            build_context_param(params, graph_name)
            if base_uri is not None:
                params["baseURI"] = base_uri
            response = self.http_client.put(
                f"/repositories/{self.identifier}/statements",
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()

    def delete(
        self,
        subj: SubjectType = None,
        pred: PredicateType = None,
        obj: ObjectType = None,
        graph_name: URIRef | Iterable[URIRef] | str | None = None,
    ) -> None:
        """Deletes statements from the repository matching the filtering parameters.

        !!! Note
            The terms for `subj`, `pred`, `obj` or `graph_name` cannot be
            [`BNodes`][rdflib.term.BNode].

        Parameters:
            subj: Subject of the statement to filter by, or `None` to match all.
            pred: Predicate of the statement to filter by, or `None` to match all.
            obj: Object of the statement to filter by, or `None` to match all.
            graph_name: Graph name(s) to restrict to.

                The default value `None` queries all graphs.

                To query just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].
        """
        validate_no_bnodes(subj, pred, obj, graph_name)
        params: dict[str, str] = {}
        build_context_param(params, graph_name)
        build_spo_param(params, subj, pred, obj)

        response = self.http_client.delete(
            f"/repositories/{self.identifier}/statements",
            params=params,
        )
        response.raise_for_status()

    @contextlib.contextmanager
    def transaction(self):
        """Create a new transaction for the repository.

        !!! warning

        Transaction instances are not thread-safe. Do not share a single
        Transaction instance across multiple threads. Each thread should create
        its own transaction, or use appropriate synchronization if sharing is
        required.
        """
        with Transaction.create(self) as txn:
            yield txn


class Transaction:
    """An RDF4J transaction.

    !!! warning

        Transaction instances are not thread-safe. Do not share a single
        Transaction instance across multiple threads. Each thread should create
        its own transaction, or use appropriate synchronization if sharing is
        required.

    Parameters:
        repo: The repository instance.
    """

    def __init__(self, repo: Repository, url: str):
        self._repo = repo
        self._url: str = url
        self._closed: bool = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.is_closed:
            if exc_type is None:
                self.commit()
            else:
                try:
                    self.rollback()
                except Exception:
                    pass

        # Propagate errors.
        return False

    @property
    def repo(self):
        """The repository instance."""
        return self._repo

    @property
    def url(self):
        """The transaction URL."""
        return self._url

    @property
    def is_closed(self) -> bool:
        """Whether the transaction is closed."""
        return self._closed

    def _raise_for_closed(self):
        if self.is_closed:
            raise TransactionClosedError("The transaction has been closed.")

    @classmethod
    def create(cls, repo: Repository) -> Transaction:
        """Create a new transaction for the repository.

        Parameters:
            repo: The repository instance.

        Returns:
            A new Transaction instance.
        """
        response = repo.http_client.post(
            f"/repositories/{repo.identifier}/transactions"
        )
        response.raise_for_status()
        url = response.headers["Location"]
        return cls(repo, url)

    def commit(self):
        """Commit the transaction.

        Raises:
            TransactionCommitError: If the transaction commit fails.
            TransactionClosedError: If the transaction is closed.
        """
        self._raise_for_closed()
        params = {"action": "COMMIT"}
        response = self.repo.http_client.put(self.url, params=params)
        if response.status_code != 200:
            raise TransactionCommitError(
                f"Transaction commit failed: {response.status_code} - {response.text}"
            )
        self._closed = True

    def rollback(self):
        """Roll back the transaction.

        Raises:
            TransactionRollbackError: If the transaction rollback fails.
            TransactionClosedError: If the transaction is closed.
        """
        self._raise_for_closed()
        response = self.repo.http_client.delete(self.url)
        if response.status_code != 204:
            raise TransactionRollbackError(
                f"Transaction rollback failed: {response.status_code} - {response.text}"
            )
        self._closed = True

    def ping(self):
        """Ping the transaction.

        Raises:
            RepositoryTransactionPingError: If the transaction ping fails.
            TransactionClosedError: If the transaction is closed.
        """
        self._raise_for_closed()
        params = {"action": "PING"}
        response = self.repo.http_client.put(self.url, params=params)
        if response.status_code != 200:
            raise TransactionPingError(
                f"Transaction ping failed: {response.status_code} - {response.text}"
            )

    def size(self, graph_name: URIRef | Iterable[URIRef] | str | None = None):
        """The number of statements in the repository or in the specified graph name.

        Parameters:
            graph_name: Graph name(s) to restrict to.

                The default value `None` queries all graphs.

                To query just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

        Returns:
            The number of statements.

        Raises:
            RepositoryResponseFormatError: Fails to parse the repository size.
        """
        self._raise_for_closed()
        validate_graph_name(graph_name)
        params = {"action": "SIZE"}
        build_context_param(params, graph_name)
        response = self.repo.http_client.put(self.url, params=params)
        response.raise_for_status()
        return self.repo._to_size(response.text)

    def query(self, query: str, **kwargs):
        """Execute a SPARQL query against the repository.

        Parameters:
            query: The SPARQL query to execute.
            **kwargs: Additional keyword arguments to include as query parameters
                in the request. See
                [RDF4J REST API - Execute SPARQL query](https://rdf4j.org/documentation/reference/rest-api/#tag/SPARQL/paths/~1repositories~1%7BrepositoryID%7D/get)
                for the list of supported query parameters.
        """
        self._raise_for_closed()
        headers: dict[str, str] = {}
        build_sparql_query_accept_header(query, headers)
        params = {"action": "QUERY", "query": query}
        response = self.repo.http_client.put(
            self.url, headers=headers, params={**params, **kwargs}
        )
        response.raise_for_status()
        try:
            return Result.parse(
                io.BytesIO(response.content),
                content_type=response.headers["Content-Type"].split(";")[0],
            )
        except KeyError as err:
            raise RDFLibParserError(
                f"Failed to parse SPARQL query result {response.headers.get('Content-Type')}: {err}"
            ) from err

    def update(self, query: str, **kwargs):
        """Execute a SPARQL update operation on the repository.

        Parameters:
            query: The SPARQL update query to execute.
            **kwargs: Additional keyword arguments to include as query parameters
                See [RDF4J REST API - Execute a transaction action](https://rdf4j.org/documentation/reference/rest-api/#tag/Transactions/paths/~1repositories~1%7BrepositoryID%7D~1transactions~1%7BtransactionID%7D/put)
                for the list of supported query parameters.
        """
        self._raise_for_closed()
        params = {"action": "UPDATE", "update": query}
        response = self.repo.http_client.put(
            self.url,
            params={**params, **kwargs},
        )
        response.raise_for_status()

    def upload(
        self,
        data: str | bytes | BinaryIO | Graph | Dataset,
        base_uri: str | None = None,
        content_type: str | None = None,
    ):
        """Upload and append statements to the repository.

        Parameters:
            data: The RDF data to upload.
            base_uri: The base URI to resolve against for any relative URIs in the data.
            content_type: The content type of the data. Defaults to
                `application/n-quads` when the value is `None`.
        """
        self._raise_for_closed()
        stream, should_close = rdf_payload_to_stream(data)
        headers = {"Content-Type": content_type or "application/n-quads"}
        params = {"action": "ADD"}
        if base_uri is not None:
            params["baseURI"] = base_uri
        try:
            response = self.repo.http_client.put(
                self.url,
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()

    def get(
        self,
        subj: SubjectType = None,
        pred: PredicateType = None,
        obj: ObjectType = None,
        graph_name: URIRef | Iterable[URIRef] | str | None = None,
        infer: bool = True,
        content_type: str | None = None,
    ) -> Graph | Dataset:
        """Get RDF statements from the repository matching the filtering parameters.

        !!! Note
            The terms for `subj`, `pred`, `obj` or `graph_name` cannot be
            [`BNodes`][rdflib.term.BNode].

        Parameters:
            subj: Subject of the statement to filter by, or `None` to match all.
            pred: Predicate of the statement to filter by, or `None` to match all.
            obj: Object of the statement to filter by, or `None` to match all.
            graph_name: Graph name(s) to restrict to.

                The default value `None` queries all graphs.

                To query just the default graph, use
                [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID].

            infer: Specifies whether inferred statements should be included in the
                result.
            content_type: The content type of the response.
                A triple-based format returns a [Graph][rdflib.graph.Graph], while a
                quad-based format returns a [`Dataset`][rdflib.graph.Dataset].

        Returns:
            A [`Graph`][rdflib.graph.Graph] or [`Dataset`][rdflib.graph.Dataset] object
                with the repository namespace prefixes bound to it.
        """
        self._raise_for_closed()
        validate_no_bnodes(subj, pred, obj, graph_name)
        if content_type is None:
            content_type = "application/n-quads"
        headers = {"Accept": content_type}
        params: dict[str, str] = {"action": "GET"}
        build_context_param(params, graph_name)
        build_spo_param(params, subj, pred, obj)
        build_infer_param(params, infer=infer)

        response = self.repo.http_client.put(
            self.url,
            headers=headers,
            params=params,
        )
        response.raise_for_status()
        triple_formats = [
            "application/n-triples",
            "text/turtle",
            "application/rdf+xml",
        ]
        try:
            if content_type in triple_formats:
                retval = Graph().parse(data=response.text, format=content_type)
            else:
                retval = Dataset().parse(data=response.text, format=content_type)
            for result in self.repo.namespaces.list():
                retval.bind(result.prefix, result.namespace, replace=True)
            return retval
        except Exception as err:
            raise RDFLibParserError(f"Error parsing RDF: {err}") from err

    def delete(
        self,
        data: str | bytes | BinaryIO | Graph | Dataset,
        base_uri: str | None = None,
        content_type: str | None = None,
    ) -> None:
        """Delete statements from the repository.

        !!! Note
            This function operates differently to
            [`Repository.delete`][rdflib.contrib.rdf4j.client.Repository.delete]
            as it does not use filter parameters. Instead, it expects a data payload.
            See the notes from
            [graphdb.js#Deleting](https://github.com/Ontotext-AD/graphdb.js?tab=readme-ov-file#deleting-1)
            for more information.

        Parameters:
            data: The RDF data to upload.
            base_uri: The base URI to resolve against for any relative URIs in the data.
            content_type: The content type of the data. Defaults to
                `application/n-quads` when the value is `None`.
        """
        self._raise_for_closed()
        params: dict[str, str] = {"action": "DELETE"}
        stream, should_close = rdf_payload_to_stream(data)
        headers = {"Content-Type": content_type or "application/n-quads"}
        if base_uri is not None:
            params["baseURI"] = base_uri
        try:
            response = self.repo.http_client.put(
                self.url,
                headers=headers,
                params=params,
                content=stream,
            )
            response.raise_for_status()
        finally:
            if should_close:
                stream.close()


class RepositoryManager:
    """A client to manage server-level repository operations.

    Parameters:
        http_client: The httpx.Client instance.
    """

    def __init__(self, http_client: httpx.Client):
        self._http_client = http_client

    @property
    def http_client(self):
        return self._http_client

    def list(self) -> list[RepositoryListingResult]:
        """List all available repositories.

        Returns:
            list[RepositoryListingResult]: List of repository results.

        Raises:
            RepositoryResponseFormatError: If the response format is unrecognized.
        """
        headers = {
            "Accept": "application/sparql-results+json",
        }
        response = self.http_client.get("/repositories", headers=headers)
        response.raise_for_status()

        try:
            data = response.json()
            results = data["results"]["bindings"]
            return [
                RepositoryListingResult(
                    identifier=repo["id"]["value"],
                    uri=repo["uri"]["value"],
                    readable=repo["readable"]["value"],
                    writable=repo["writable"]["value"],
                    title=repo.get("title", {}).get("value"),
                )
                for repo in results
            ]
        except (KeyError, ValueError) as err:
            raise RepositoryResponseFormatError(f"Unrecognised response format: {err}")

    def get(self, repository_id: str) -> Repository:
        """Get a repository by ID.

        !!! Note
            This performs a health check before returning the repository object.

        Parameters:
            repository_id: The identifier of the repository.

        Returns:
            Repository: The repository instance.

        Raises:
            RepositoryNotFoundError: If the repository is not found.
            RepositoryNotHealthyError: If the repository is not healthy.
        """
        repo = Repository(repository_id, self.http_client)
        repo.health()
        return repo

    def create(
        self, repository_id: str, data: str, content_type: str = "text/turtle"
    ) -> Repository:
        """Create a new repository.

        Parameters:
            repository_id: The identifier of the repository.
            data: The repository configuration in RDF.
            content_type: The repository configuration content type.

        Raises:
            RepositoryAlreadyExistsError: If the repository already exists.
            RepositoryNotHealthyError: If the repository is not healthy.
        """
        try:
            headers = {"Content-Type": content_type}
            response = self.http_client.put(
                f"/repositories/{repository_id}", headers=headers, content=data
            )
            response.raise_for_status()
            return self.get(repository_id)
        except httpx.HTTPStatusError as err:
            if err.response.status_code == 409:
                raise RepositoryAlreadyExistsError(
                    f"Repository {repository_id} already exists."
                )
            raise

    def delete(self, repository_id: str) -> None:
        """Delete a repository.

        Parameters:
            repository_id: The identifier of the repository.

        Raises:
            RepositoryNotFoundError: If the repository is not found.
            RepositoryError: If the repository is not deleted successfully.
        """
        try:
            response = self.http_client.delete(f"/repositories/{repository_id}")
            response.raise_for_status()
            if response.status_code != 204:
                raise RepositoryError(
                    f"Unexpected response status code when deleting repository {repository_id}: {response.status_code} - {response.text.strip()}"
                )
        except httpx.HTTPStatusError as err:
            if err.response.status_code == 404:
                raise RepositoryNotFoundError(f"Repository {repository_id} not found.")
            raise


class RDF4JClient:
    """RDF4J client.

    This client and its inner management objects perform HTTP requests via
    httpx and may raise httpx-specific exceptions. Errors documented by RDF4J
    in its protocol specification are mapped to specific exceptions in this
    library where applicable. Error mappings are documented on each management
    method. The underlying httpx client is reused across requests, and
    connection pooling is handled automatically by httpx.

    Parameters:
        base_url: The base URL of the RDF4J server.
        auth: Authentication credentials. Can be a tuple (username, password) for
            basic auth, or a string for token-based auth (e.g., "GDB <token>")
            which is added as the Authorization header.
        timeout: Request timeout in seconds or an httpx.Timeout for fine-grained control (default: 30.0).
        kwargs: Additional keyword arguments to pass to the httpx.Client.
    """

    def __init__(
        self,
        base_url: str,
        auth: tuple[str, str] | str | None = None,
        timeout: float | httpx.Timeout = 30.0,
        **kwargs: Any,
    ):
        if not base_url.endswith("/"):
            base_url += "/"

        httpx_auth: tuple[str, str] | None = None
        if isinstance(auth, tuple):
            httpx_auth = auth
        elif isinstance(auth, str):
            headers = kwargs.get("headers", {})
            headers["Authorization"] = auth
            kwargs["headers"] = headers

        self._http_client = httpx.Client(
            base_url=base_url, auth=httpx_auth, timeout=timeout, **kwargs
        )
        self._repository_manager = RepositoryManager(self.http_client)
        try:
            protocol_version = self.protocol
        except httpx.RequestError as err:
            self.close()
            raise RDF4JUnsupportedProtocolError(
                f"Failed to check protocol version: {err}"
            ) from err
        if protocol_version < 12:
            self.close()
            raise RDF4JUnsupportedProtocolError(
                f"RDF4J server protocol version {protocol_version} is not supported. Minimum required version is 12."
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def http_client(self):
        return self._http_client

    @property
    def repositories(self) -> RepositoryManager:
        """Server-level repository management operations."""
        return self._repository_manager

    @property
    def protocol(self) -> float:
        """The RDF4J REST API protocol version.

        Returns:
            The protocol version number.
        """
        response = self.http_client.get("/protocol", headers={"Accept": "text/plain"})
        response.raise_for_status()
        return float(response.text.strip())

    def close(self):
        """Close the underlying httpx.Client."""
        self.http_client.close()



"""Plugin support for rdf.

There are a number of plugin points for rdf: parser, serializer,
store, query processor, and query result. Plugins can be registered
either through setuptools entry_points or by calling
rdf.plugin.register directly.

If you have a package that uses a setuptools based `setup.py` you can add the
following to your setup:

```python
entry_points = {
    'rdf.plugins.parser': [
        'nt =     rdf.plugins.parsers.ntriples:NTParser',
        ],
    'rdf.plugins.serializer': [
        'nt =     rdf.plugins.serializers.NTSerializer:NTSerializer',
        ],
    }
```

See the [setuptools dynamic discovery of services and plugins](http://peak.telecommunity.com/DevCenter/setuptools#dynamic-discovery-of-services-and-plugins)
for moreinformation.
"""

from __future__ import annotations

from importlib.metadata import EntryPoint, entry_points
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    overload,
)

import rdflib.plugins.stores.berkeleydb
from rdflib.exceptions import Error
from rdflib.parser import Parser
from rdflib.query import (
    Processor,
    Result,
    ResultParser,
    ResultSerializer,
    UpdateProcessor,
)
from rdflib.serializer import Serializer
from rdflib.store import Store

__all__ = [
    "register",
    "get",
    "plugins",
    "PluginException",
    "Plugin",
    "PluginT",
    "PKGPlugin",
]

rdflib_entry_points = {
    "rdf.plugins.store": Store,
    "rdf.plugins.serializer": Serializer,
    "rdf.plugins.parser": Parser,
    "rdf.plugins.resultparser": ResultParser,
    "rdf.plugins.resultserializer": ResultSerializer,
    "rdf.plugins.queryprocessor": Processor,
    "rdf.plugins.queryresult": Result,
    "rdf.plugins.updateprocessor": UpdateProcessor,
}

_plugins: Dict[Tuple[str, Type[Any]], Plugin] = {}


class PluginException(Error):  # noqa: N818
    pass


#: A generic type variable for plugins
PluginT = TypeVar("PluginT")


class Plugin(Generic[PluginT]):
    def __init__(
        self, name: str, kind: Type[PluginT], module_path: str, class_name: str
    ):
        self.name = name
        self.kind = kind
        self.module_path = module_path
        self.class_name = class_name
        self._class: Optional[Type[PluginT]] = None

    def getClass(self) -> Type[PluginT]:  # noqa: N802
        if self._class is None:
            module = __import__(self.module_path, globals(), locals(), [""])
            self._class = getattr(module, self.class_name)
        return self._class


class PKGPlugin(Plugin[PluginT]):
    def __init__(self, name: str, kind: Type[PluginT], ep: EntryPoint):
        self.name = name
        self.kind = kind
        self.ep = ep
        self._class: Optional[Type[PluginT]] = None

    def getClass(self) -> Type[PluginT]:  # noqa: N802
        if self._class is None:
            self._class = self.ep.load()
        return self._class


def register(name: str, kind: Type[Any], module_path, class_name):
    """
    Register the plugin for (name, kind). The module_path and
    class_name should be the path to a plugin class.
    """
    p = Plugin(name, kind, module_path, class_name)
    _plugins[(name, kind)] = p


def get(name: str, kind: Type[PluginT]) -> Type[PluginT]:
    """
    Return the class for the specified (name, kind). Raises a
    PluginException if unable to do so.
    """
    try:
        p: Plugin[PluginT] = _plugins[(name, kind)]
    except KeyError:
        raise PluginException("No plugin registered for (%s, %s)" % (name, kind))
    return p.getClass()


all_entry_points = entry_points()
if hasattr(all_entry_points, "select"):
    for entry_point, kind in rdflib_entry_points.items():
        for ep in all_entry_points.select(group=entry_point):
            _plugins[(ep.name, kind)] = PKGPlugin(ep.name, kind, ep)
else:
    # Prior to Python 3.10, this returns a dict instead of the selection interface, which is slightly slower
    if TYPE_CHECKING:
        assert isinstance(all_entry_points, dict)
    for entry_point, kind in rdflib_entry_points.items():
        for ep in all_entry_points.get(entry_point, []):
            _plugins[(ep.name, kind)] = PKGPlugin(ep.name, kind, ep)


@overload
def plugins(
    name: Optional[str] = ..., kind: Type[PluginT] = ...
) -> Iterator[Plugin[PluginT]]: ...


@overload
def plugins(name: Optional[str] = ..., kind: None = ...) -> Iterator[Plugin]: ...


def plugins(
    name: Optional[str] = None, kind: Optional[Type[PluginT]] = None
) -> Iterator[Plugin[PluginT]]:
    """
    A generator of the plugins.

    Pass in name and kind to filter... else leave None to match all.
    """
    for p in _plugins.values():
        if (name is None or name == p.name) and (kind is None or kind == p.kind):
            yield p


# Register Stores

if rdflib.plugins.stores.berkeleydb.has_bsddb:
    # Checks for BerkeleyDB before registering it
    register(
        "BerkeleyDB",
        Store,
        "rdflib.plugins.stores.berkeleydb",
        "BerkeleyDB",
    )
register(
    "default",
    Store,
    "rdflib.plugins.stores.memory",
    "Memory",
)
register(
    "Memory",
    Store,
    "rdflib.plugins.stores.memory",
    "Memory",
)
register(
    "SimpleMemory",
    Store,
    "rdflib.plugins.stores.memory",
    "SimpleMemory",
)
register(
    "Auditable",
    Store,
    "rdflib.plugins.stores.auditable",
    "AuditableStore",
)
register(
    "Concurrent",
    Store,
    "rdflib.plugins.stores.concurrent",
    "ConcurrentStore",
)

register(
    "SPARQLStore",
    Store,
    "rdflib.plugins.stores.sparqlstore",
    "SPARQLStore",
)
register(
    "SPARQLUpdateStore",
    Store,
    "rdflib.plugins.stores.sparqlstore",
    "SPARQLUpdateStore",
)

# Register Triple Serializers
register(
    "application/rdf+xml",
    Serializer,
    "rdflib.plugins.serializers.rdfxml",
    "XMLSerializer",
)
register(
    "xml",
    Serializer,
    "rdflib.plugins.serializers.rdfxml",
    "XMLSerializer",
)
register(
    "pretty-xml",
    Serializer,
    "rdflib.plugins.serializers.rdfxml",
    "PrettyXMLSerializer",
)
register(
    "text/n3",
    Serializer,
    "rdflib.plugins.serializers.n3",
    "N3Serializer",
)
register(
    "n3",
    Serializer,
    "rdflib.plugins.serializers.n3",
    "N3Serializer",
)
register(
    "text/turtle",
    Serializer,
    "rdflib.plugins.serializers.turtle",
    "TurtleSerializer",
)
register(
    "turtle",
    Serializer,
    "rdflib.plugins.serializers.turtle",
    "TurtleSerializer",
)
register(
    "ttl",
    Serializer,
    "rdflib.plugins.serializers.turtle",
    "TurtleSerializer",
)
register(
    "longturtle",
    Serializer,
    "rdflib.plugins.serializers.longturtle",
    "LongTurtleSerializer",
)
register(
    "application/n-triples",
    Serializer,
    "rdflib.plugins.serializers.nt",
    "NTSerializer",
)
register(
    "ntriples",
    Serializer,
    "rdflib.plugins.serializers.nt",
    "NTSerializer",
)
register(
    "nt",
    Serializer,
    "rdflib.plugins.serializers.nt",
    "NTSerializer",
)
register(
    "nt11",
    Serializer,
    "rdflib.plugins.serializers.nt",
    "NT11Serializer",
)
register(
    "json-ld",
    Serializer,
    "rdflib.plugins.serializers.jsonld",
    "JsonLDSerializer",
)
register(
    "application/ld+json",
    Serializer,
    "rdflib.plugins.serializers.jsonld",
    "JsonLDSerializer",
)

# Register Quad Serializers
register(
    "application/n-quads",
    Serializer,
    "rdflib.plugins.serializers.nquads",
    "NQuadsSerializer",
)
register(
    "nquads",
    Serializer,
    "rdflib.plugins.serializers.nquads",
    "NQuadsSerializer",
)
register(
    "application/trix",
    Serializer,
    "rdflib.plugins.serializers.trix",
    "TriXSerializer",
)
register(
    "trix",
    Serializer,
    "rdflib.plugins.serializers.trix",
    "TriXSerializer",
)
register(
    "application/trig",
    Serializer,
    "rdflib.plugins.serializers.trig",
    "TrigSerializer",
)
register(
    "trig",
    Serializer,
    "rdflib.plugins.serializers.trig",
    "TrigSerializer",
)
register(
    "hext",
    Serializer,
    "rdflib.plugins.serializers.hext",
    "HextuplesSerializer",
)
register(
    "patch",
    Serializer,
    "rdflib.plugins.serializers.patch",
    "PatchSerializer",
)

# Register Triple Parsers
register(
    "application/rdf+xml",
    Parser,
    "rdflib.plugins.parsers.rdfxml",
    "RDFXMLParser",
)
register(
    "xml",
    Parser,
    "rdflib.plugins.parsers.rdfxml",
    "RDFXMLParser",
)
register(
    "text/n3",
    Parser,
    "rdflib.plugins.parsers.notation3",
    "N3Parser",
)
register(
    "n3",
    Parser,
    "rdflib.plugins.parsers.notation3",
    "N3Parser",
)
register(
    "text/turtle",
    Parser,
    "rdflib.plugins.parsers.notation3",
    "TurtleParser",
)
register(
    "turtle",
    Parser,
    "rdflib.plugins.parsers.notation3",
    "TurtleParser",
)
register(
    "ttl",
    Parser,
    "rdflib.plugins.parsers.notation3",
    "TurtleParser",
)
register(
    "application/n-triples",
    Parser,
    "rdflib.plugins.parsers.ntriples",
    "NTParser",
)
register(
    "ntriples",
    Parser,
    "rdflib.plugins.parsers.ntriples",
    "NTParser",
)
register(
    "nt",
    Parser,
    "rdflib.plugins.parsers.ntriples",
    "NTParser",
)
register(
    "nt11",
    Parser,
    "rdflib.plugins.parsers.ntriples",
    "NTParser",
)
register(
    "application/ld+json",
    Parser,
    "rdflib.plugins.parsers.jsonld",
    "JsonLDParser",
)
register(
    "json-ld",
    Parser,
    "rdflib.plugins.parsers.jsonld",
    "JsonLDParser",
)

# Register Quad Parsers
register(
    "application/n-quads",
    Parser,
    "rdflib.plugins.parsers.nquads",
    "NQuadsParser",
)
register(
    "nquads",
    Parser,
    "rdflib.plugins.parsers.nquads",
    "NQuadsParser",
)
register(
    "application/trix",
    Parser,
    "rdflib.plugins.parsers.trix",
    "TriXParser",
)
register(
    "trix",
    Parser,
    "rdflib.plugins.parsers.trix",
    "TriXParser",
)
register(
    "application/trig",
    Parser,
    "rdflib.plugins.parsers.trig",
    "TrigParser",
)
register(
    "trig",
    Parser,
    "rdflib.plugins.parsers.trig",
    "TrigParser",
)
register(
    "hext",
    Parser,
    "rdflib.plugins.parsers.hext",
    "HextuplesParser",
)

# Register RDF Patch Parsers
register(
    "patch",
    Parser,
    "rdflib.plugins.parsers.patch",
    "RDFPatchParser",
)

# Register SPARQL Processors
register(
    "sparql",
    Result,
    "rdflib.plugins.sparql.processor",
    "SPARQLResult",
)
register(
    "sparql",
    Processor,
    "rdflib.plugins.sparql.processor",
    "SPARQLProcessor",
)
register(
    "sparql",
    UpdateProcessor,
    "rdflib.plugins.sparql.processor",
    "SPARQLUpdateProcessor",
)

# Register SPARQL Result Serializers
register(
    "xml",
    ResultSerializer,
    "rdflib.plugins.sparql.results.xmlresults",
    "XMLResultSerializer",
)
register(
    "application/sparql-results+xml",
    ResultSerializer,
    "rdflib.plugins.sparql.results.xmlresults",
    "XMLResultSerializer",
)
register(
    "txt",
    ResultSerializer,
    "rdflib.plugins.sparql.results.txtresults",
    "TXTResultSerializer",
)
register(
    "json",
    ResultSerializer,
    "rdflib.plugins.sparql.results.jsonresults",
    "JSONResultSerializer",
)
register(
    "application/sparql-results+json",
    ResultSerializer,
    "rdflib.plugins.sparql.results.jsonresults",
    "JSONResultSerializer",
)
register(
    "csv",
    ResultSerializer,
    "rdflib.plugins.sparql.results.csvresults",
    "CSVResultSerializer",
)
register(
    "text/csv",
    ResultSerializer,
    "rdflib.plugins.sparql.results.csvresults",
    "CSVResultSerializer",
)

# Register SPARQL Result Parsers
register(
    "xml",
    ResultParser,
    "rdflib.plugins.sparql.results.xmlresults",
    "XMLResultParser",
)
register(
    "application/sparql-results+xml",
    ResultParser,
    "rdflib.plugins.sparql.results.xmlresults",
    "XMLResultParser",
)
register(
    "json",
    ResultParser,
    "rdflib.plugins.sparql.results.jsonresults",
    "JSONResultParser",
)
register(
    "application/sparql-results+json",
    ResultParser,
    "rdflib.plugins.sparql.results.jsonresults",
    "JSONResultParser",
)
register(
    "csv",
    ResultParser,
    "rdflib.plugins.sparql.results.csvresults",
    "CSVResultParser",
)
register(
    "text/csv",
    ResultParser,
    "rdflib.plugins.sparql.results.csvresults",
    "CSVResultParser",
)
register(
    "tsv",
    ResultParser,
    "rdflib.plugins.sparql.results.tsvresults",
    "TSVResultParser",
)
register(
    "text/tab-separated-values",
    ResultParser,
    "rdflib.plugins.sparql.results.tsvresults",
    "TSVResultParser",
)

graph_parsers = {parser.name for parser in plugins(kind=Parser)}
result_parsers = {parser.name for parser in plugins(kind=ResultParser)}
graph_result_parsers = graph_parsers - result_parsers
for parser_name in graph_result_parsers:
    register(
        parser_name,
        ResultParser,
        "rdflib.plugins.sparql.results.graph",
        "GraphResultParser",
    )



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class SSN(DefinedNamespace):
    """
    Semantic Sensor Network Ontology

    This ontology describes sensors, actuators and observations, and related concepts. It does not describe domain
    concepts, time, locations, etc. these are intended to be included from other ontologies via OWL imports.

    Generated from: http://www.w3.org/ns/ssn/
    Date: 2020-05-26 14:20:09.068204

    """

    # http://www.w3.org/2002/07/owl#Class
    Deployment: URIRef  # Describes the Deployment of one or more Systems for a particular purpose. Deployment may be done on a Platform.
    Input: URIRef  # Any information that is provided to a Procedure for its use.
    Output: URIRef  # Any information that is reported from a Procedure.
    Property: URIRef  # A quality of an entity. An aspect of an entity that is intrinsic to and cannot exist without the entity.
    Stimulus: URIRef  # An event in the real world that 'triggers' the Sensor. The properties associated to the Stimulus may be different to the eventual observed ObservableProperty. It is the event, not the object, that triggers the Sensor.
    System: URIRef  # System is a unit of abstraction for pieces of infrastructure that implement Procedures. A System may have components, its subsystems, which are other systems.

    # http://www.w3.org/2002/07/owl#FunctionalProperty
    wasOriginatedBy: (
        URIRef  # Relation between an Observation and the Stimulus that originated it.
    )

    # http://www.w3.org/2002/07/owl#ObjectProperty
    deployedOnPlatform: URIRef  # Relation between a Deployment and the Platform on which the Systems are deployed.
    deployedSystem: URIRef  # Relation between a Deployment and a deployed System.
    detects: URIRef  # A relation from a Sensor to the Stimulus that the Sensor detects. The Stimulus itself will be serving as a proxy for some ObservableProperty.
    forProperty: URIRef  # A relation between some aspect of an entity and a Property.
    hasDeployment: URIRef  # Relation between a System and a Deployment, recording that the System is deployed in that Deployment.
    hasInput: URIRef  # Relation between a Procedure and an Input to it.
    hasOutput: URIRef  # Relation between a Procedure and an Output of it.
    hasProperty: URIRef  # Relation between an entity and a Property of that entity.
    hasSubSystem: URIRef  # Relation between a System and its component parts.
    implementedBy: URIRef  # Relation between a Procedure (an algorithm, procedure or method) and an entity that implements that Procedure in some executable way.
    implements: URIRef  # Relation between an entity that implements a Procedure in some executable way and the Procedure (an algorithm, procedure or method).
    inDeployment: URIRef  # Relation between a Platform and a Deployment, meaning that the deployedSystems of the Deployment are hosted on the Platform.
    isPropertyOf: URIRef  # Relation between a Property and the entity it belongs to.
    isProxyFor: URIRef  # A relation from a Stimulus to the Property that the Stimulus is serving as a proxy for.

    _NS = Namespace("http://www.w3.org/ns/ssn/")



"""
LongTurtle RDF graph serializer for RDFLib.
See http://www.w3.org/TeamSubmission/turtle/ for syntax specification.

This variant, longturtle as opposed to just turtle, makes some small format changes
to turtle - the original turtle serializer. It:

* uses PREFIX instead of @prefix
* uses BASE instead of @base
* adds a new line at RDF.type, or 'a'
* adds a newline and an indent for all triples with more than one object (object list)
* adds a new line and ';' for the last triple in a set with '.'
    on the start of the next line
* uses default encoding (encode()) is used instead of "latin-1"

- Nicholas Car, 2023
"""

from __future__ import annotations

import warnings
from typing import IO, Any, Optional

from rdflib.compare import to_canonical_graph
from rdflib.exceptions import Error
from rdflib.graph import Graph, _TripleType
from rdflib.namespace import RDF
from rdflib.term import BNode, Literal, URIRef

from .turtle import RecursiveSerializer

__all__ = ["LongTurtleSerializer"]

SUBJECT = 0
VERB = 1
OBJECT = 2

_GEN_QNAME_FOR_DT = False
_SPACIOUS_OUTPUT = False


class LongTurtleSerializer(RecursiveSerializer):
    """LongTurtle, a Turtle serialization format.

    When the optional parameter `canon` is set to `True`, the graph is canonicalized
    before serialization. This normalizes blank node identifiers and allows for
    deterministic serialization of the graph. Useful when consistent outputs are required.
    """

    short_name = "longturtle"
    indentString = "    "

    def __init__(self, store):
        self._ns_rewrite = {}
        self._canon = False
        super(LongTurtleSerializer, self).__init__(store)
        self.keywords = {RDF.type: "a"}
        self.reset()
        self.stream = None
        self._spacious: bool = _SPACIOUS_OUTPUT

    def addNamespace(self, prefix, namespace):
        # Turtle does not support prefixes that start with _
        # if they occur in the graph, rewrite to p_blah
        # this is more complicated since we need to make sure p_blah
        # does not already exist. And we register namespaces as we go, i.e.
        # we may first see a triple with prefix _9 - rewrite it to p_9
        # and then later find a triple with a "real" p_9 prefix

        # so we need to keep track of ns rewrites we made so far.

        if (prefix > "" and prefix[0] == "_") or self.namespaces.get(
            prefix, namespace
        ) != namespace:
            if prefix not in self._ns_rewrite:
                p = "p" + prefix
                while p in self.namespaces:
                    p = "p" + p
                self._ns_rewrite[prefix] = p

            prefix = self._ns_rewrite.get(prefix, prefix)

        super(LongTurtleSerializer, self).addNamespace(prefix, namespace)
        return prefix

    def canonize(self):
        """Apply canonicalization to the store.

        This normalizes blank node identifiers and allows for deterministic
        serialization of the graph.
        """
        if not self._canon:
            return

        namespace_manager = self.store.namespace_manager
        store = to_canonical_graph(self.store)
        content = store.serialize(format="application/n-triples")
        lines = content.split("\n")
        lines.sort()
        graph = Graph()
        graph.parse(
            data="\n".join(lines), format="application/n-triples", skolemize=True
        )
        graph = graph.de_skolemize()
        graph.namespace_manager = namespace_manager
        self.store = graph

    def reset(self):
        super(LongTurtleSerializer, self).reset()
        self._shortNames = {}
        self._started = False
        self._ns_rewrite = {}
        self.canonize()

    def serialize(
        self,
        stream: IO[bytes],
        base: Optional[str] = None,
        encoding: Optional[str] = None,
        spacious: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        self._canon = kwargs.get("canon", False)
        self.reset()
        self.stream = stream
        # if base is given here, use, if not and a base is set for the graph use that
        if base is not None:
            self.base = base
        elif self.store.base is not None:
            self.base = self.store.base

        if spacious is not None:
            self._spacious = spacious

        self.preprocess()
        subjects_list = self.orderSubjects()

        self.startDocument()

        firstTime = True
        for subject in subjects_list:
            if self.isDone(subject):
                continue
            if firstTime:
                firstTime = False
            if self.statement(subject) and not firstTime:
                self.write("\n")

        self.endDocument()

        self.base = None

    def preprocessTriple(self, triple: _TripleType) -> None:
        super(LongTurtleSerializer, self).preprocessTriple(triple)
        for i, node in enumerate(triple):
            if i == VERB:
                if node in self.keywords:
                    # predicate is a keyword
                    continue
                if (
                    self.base is not None
                    and isinstance(node, URIRef)
                    and node.startswith(self.base)
                    and "#" not in node.replace(self.base, "")
                    and "/" not in node.replace(self.base, "")
                ):
                    # predicate corresponds to base namespace
                    continue
            # Don't use generated prefixes for subjects and objects
            self.get_pname(node, gen_prefix=(i == VERB))
            if isinstance(node, Literal) and node.datatype:
                self.get_pname(node.datatype, gen_prefix=_GEN_QNAME_FOR_DT)
        p = triple[1]
        if isinstance(p, BNode):  # hmm - when is P ever a bnode?
            self._references[p] += 1

    def get_pname(self, uri, gen_prefix=True):
        if not isinstance(uri, URIRef):
            return None

        try:
            parts = self.store.compute_qname(uri, generate=gen_prefix)
        except Exception:
            # is the uri a namespace in itself?
            pfx = self.store.store.prefix(uri)

            if pfx is not None:
                parts = (pfx, uri, "")
            else:
                # nothing worked
                return None

        prefix, namespace, local = parts

        # To understand treatment of % character refer to Productions for terminal PLX at
        # https://www.w3.org/TR/turtle/#grammar-production-PLX
        # Only % NOT followed by two hex chars requires manual backslash escaping
        local = local.replace(r"(", r"\(").replace(r")", r"\)")
        local = self.LOCALNAME_PECRENT_CHARACTER_REQUIRING_ESCAPE_REGEX.sub(
            "\\%", local
        )

        # PName cannot end with .
        if local.endswith("."):
            return None

        prefix = self.addNamespace(prefix, namespace)

        return "%s:%s" % (prefix, local)

    def getQName(self, uri, gen_prefix=True):
        warnings.warn(
            "LongTurtleSerializer.getQName is deprecated, use LongTurtleSerializer.get_pname instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_pname(uri, gen_prefix)

    def startDocument(self):
        self._started = True
        ns_list = sorted(self.namespaces.items())

        if self.base:
            self.write(self.indent() + "BASE <%s>\n" % self.base)
        for prefix, uri in ns_list:
            self.write(self.indent() + "PREFIX %s: <%s>\n" % (prefix, uri))
        if ns_list and self._spacious:
            self.write("\n")

    def endDocument(self):
        if self._spacious:
            self.write("\n")

    def statement(self, subject):
        self.subjectDone(subject)
        return self.s_squared(subject) or self.s_default(subject)

    def s_default(self, subject):
        self.write("\n" + self.indent())
        self.path(subject, SUBJECT)
        self.write("\n" + self.indent())
        self.predicateList(subject)
        self.write("\n.")
        return True

    def s_squared(self, subject):
        if (self._references[subject] > 0) or not isinstance(subject, BNode):
            return False
        self.write("\n" + self.indent() + "[]")
        self.predicateList(subject, newline=False)
        self.write("\n.")
        return True

    def path(self, node, position, newline=False):
        if not (
            self.p_squared(node, position) or self.p_default(node, position, newline)
        ):
            raise Error("Cannot serialize node '%s'" % (node,))

    def p_default(self, node, position, newline=False):
        if position != SUBJECT and not newline:
            self.write(" ")
        self.write(self.label(node, position))
        return True

    def label(self, node, position):
        if node == RDF.nil:
            return "()"
        if position is VERB and node in self.keywords:
            return self.keywords[node]
        if isinstance(node, Literal):
            return node._literal_n3(
                use_plain=True,
                qname_callback=lambda dt: self.get_pname(dt, _GEN_QNAME_FOR_DT),
            )
        else:
            node = self.relativize(node)

            return self.get_pname(node, position == VERB) or node.n3()

    def p_squared(
        self,
        node,
        position,
    ):
        if (
            not isinstance(node, BNode)
            or node in self._serialized
            or self._references[node] > 1
            or position == SUBJECT
        ):
            return False

        if self.isValidList(node):
            # this is a list
            self.depth += 2
            self.write(" (\n")
            self.depth -= 2
            self.doList(node)
            self.write("\n" + self.indent() + ")")
        else:
            # this is a Blank Node
            self.subjectDone(node)
            self.write("\n" + self.indent(1) + "[\n")
            self.depth += 1
            self.predicateList(node)
            self.depth -= 1
            self.write("\n" + self.indent(1) + "]")

        return True

    def isValidList(self, l_):
        """
        Checks if l is a valid RDF list, i.e. no nodes have other properties.
        """
        try:
            if self.store.value(l_, RDF.first) is None:
                return False
        except Exception:
            return False
        while l_:
            if l_ != RDF.nil and len(list(self.store.predicate_objects(l_))) != 2:
                return False
            l_ = self.store.value(l_, RDF.rest)
        return True

    def doList(self, l_):
        i = 0
        while l_:
            item = self.store.value(l_, RDF.first)
            if item is not None:
                if i == 0:
                    self.write(self.indent(1))
                else:
                    self.write("\n" + self.indent(1))
                self.path(item, OBJECT, newline=True)
                self.subjectDone(l_)
            l_ = self.store.value(l_, RDF.rest)
            i += 1

    def predicateList(self, subject, newline=False):
        properties = self.buildPredicateHash(subject)
        propList = self.sortProperties(properties)
        if len(propList) == 0:
            return
        self.write(self.indent(1))
        self.verb(propList[0], newline=True)
        self.objectList(properties[propList[0]])
        for predicate in propList[1:]:
            self.write(" ;\n" + self.indent(1))
            self.verb(predicate, newline=True)
            self.objectList(properties[predicate])
        self.write(" ;")

    def verb(self, node, newline=False):
        self.path(node, VERB, newline)

    def objectList(self, objects):
        count = len(objects)
        if count == 0:
            return
        depthmod = (count == 1) and 0 or 1
        self.depth += depthmod
        first_nl = False
        if count > 1:
            if not isinstance(objects[0], BNode):
                self.write("\n" + self.indent(1))
            else:
                self.write(" ")
            first_nl = True
        self.path(objects[0], OBJECT, newline=first_nl)
        for obj in objects[1:]:
            self.write(" ,")
            if not isinstance(obj, BNode):
                self.write("\n" + self.indent(1))
            self.path(obj, OBJECT, newline=True)
        self.depth -= depthmod



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class OWL(DefinedNamespace):
    """
    The OWL 2 Schema vocabulary (OWL 2)

    This ontology partially describes the built-in classes and    properties that together form the basis of
    the RDF/XML syntax of OWL 2.    The content of this ontology is based on Tables 6.1 and 6.2    in Section 6.4
    of the OWL 2 RDF-Based Semantics specification,    available at http://www.w3.org/TR/owl2-rdf-based-
    semantics/.    Please note that those tables do not include the different annotations    (labels, comments and
    rdfs:isDefinedBy links) used in this file.    Also note that the descriptions provided in this ontology do not
    provide a complete and correct formal description of either the syntax    or the semantics of the introduced
    terms (please see the OWL 2    recommendations for the complete and normative specifications).    Furthermore,
    the information provided by this ontology may be    misleading if not used with care. This ontology SHOULD NOT
    be imported    into OWL ontologies. Importing this file into an OWL 2 DL ontology    will cause it to become
    an OWL 2 Full ontology and may have other,    unexpected, consequences.

    Generated from: http://www.w3.org/2002/07/owl#
    Date: 2020-05-26 14:20:03.193795

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    allValuesFrom: URIRef  # The property that determines the class that a universal property restriction refers to.
    annotatedProperty: URIRef  # The property that determines the predicate of an annotated axiom or annotated annotation.
    annotatedSource: URIRef  # The property that determines the subject of an annotated axiom or annotated annotation.
    annotatedTarget: URIRef  # The property that determines the object of an annotated axiom or annotated annotation.
    assertionProperty: URIRef  # The property that determines the predicate of a negative property assertion.
    cardinality: URIRef  # The property that determines the cardinality of an exact cardinality restriction.
    complementOf: URIRef  # The property that determines that a given class is the complement of another class.
    datatypeComplementOf: URIRef  # The property that determines that a given data range is the complement of another data range with respect to the data domain.
    differentFrom: (
        URIRef  # The property that determines that two given individuals are different.
    )
    disjointUnionOf: URIRef  # The property that determines that a given class is equivalent to the disjoint union of a collection of other classes.
    disjointWith: (
        URIRef  # The property that determines that two given classes are disjoint.
    )
    distinctMembers: URIRef  # The property that determines the collection of pairwise different individuals in a owl:AllDifferent axiom.
    equivalentClass: URIRef  # The property that determines that two given classes are equivalent, and that is used to specify datatype definitions.
    equivalentProperty: (
        URIRef  # The property that determines that two given properties are equivalent.
    )
    hasKey: URIRef  # The property that determines the collection of properties that jointly build a key.
    hasSelf: URIRef  # The property that determines the property that a self restriction refers to.
    hasValue: URIRef  # The property that determines the individual that a has-value restriction refers to.
    intersectionOf: URIRef  # The property that determines the collection of classes or data ranges that build an intersection.
    inverseOf: (
        URIRef  # The property that determines that two given properties are inverse.
    )
    maxCardinality: URIRef  # The property that determines the cardinality of a maximum cardinality restriction.
    maxQualifiedCardinality: URIRef  # The property that determines the cardinality of a maximum qualified cardinality restriction.
    members: URIRef  # The property that determines the collection of members in either a owl:AllDifferent, owl:AllDisjointClasses or owl:AllDisjointProperties axiom.
    minCardinality: URIRef  # The property that determines the cardinality of a minimum cardinality restriction.
    minQualifiedCardinality: URIRef  # The property that determines the cardinality of a minimum qualified cardinality restriction.
    onClass: URIRef  # The property that determines the class that a qualified object cardinality restriction refers to.
    onDataRange: URIRef  # The property that determines the data range that a qualified data cardinality restriction refers to.
    onDatatype: URIRef  # The property that determines the datatype that a datatype restriction refers to.
    onProperties: URIRef  # The property that determines the n-tuple of properties that a property restriction on an n-ary data range refers to.
    onProperty: URIRef  # The property that determines the property that a property restriction refers to.
    oneOf: URIRef  # The property that determines the collection of individuals or data values that build an enumeration.
    propertyChainAxiom: URIRef  # The property that determines the n-tuple of properties that build a sub property chain of a given property.
    propertyDisjointWith: (
        URIRef  # The property that determines that two given properties are disjoint.
    )
    qualifiedCardinality: URIRef  # The property that determines the cardinality of an exact qualified cardinality restriction.
    sameAs: URIRef  # The property that determines that two given individuals are equal.
    someValuesFrom: URIRef  # The property that determines the class that an existential property restriction refers to.
    sourceIndividual: URIRef  # The property that determines the subject of a negative property assertion.
    targetIndividual: URIRef  # The property that determines the object of a negative object property assertion.
    targetValue: URIRef  # The property that determines the value of a negative data property assertion.
    unionOf: URIRef  # The property that determines the collection of classes or data ranges that build a union.
    withRestrictions: URIRef  # The property that determines the collection of facet-value pairs that define a datatype restriction.

    # http://www.w3.org/2000/01/rdf-schema#Class
    AllDifferent: URIRef  # The class of collections of pairwise different individuals.
    AllDisjointClasses: URIRef  # The class of collections of pairwise disjoint classes.
    AllDisjointProperties: (
        URIRef  # The class of collections of pairwise disjoint properties.
    )
    Annotation: URIRef  # The class of annotated annotations for which the RDF serialization consists of an annotated subject, predicate and object.
    AnnotationProperty: URIRef  # The class of annotation properties.
    AsymmetricProperty: URIRef  # The class of asymmetric properties.
    Axiom: URIRef  # The class of annotated axioms for which the RDF serialization consists of an annotated subject, predicate and object.
    Class: URIRef  # The class of OWL classes.
    DataRange: URIRef  # The class of OWL data ranges, which are special kinds of datatypes. Note: The use of the IRI owl:DataRange has been deprecated as of OWL 2. The IRI rdfs:Datatype SHOULD be used instead.
    DatatypeProperty: URIRef  # The class of data properties.
    DeprecatedClass: URIRef  # The class of deprecated classes.
    DeprecatedProperty: URIRef  # The class of deprecated properties.
    FunctionalProperty: URIRef  # The class of functional properties.
    InverseFunctionalProperty: URIRef  # The class of inverse-functional properties.
    IrreflexiveProperty: URIRef  # The class of irreflexive properties.
    NamedIndividual: URIRef  # The class of named individuals.
    NegativePropertyAssertion: URIRef  # The class of negative property assertions.
    ObjectProperty: URIRef  # The class of object properties.
    Ontology: URIRef  # The class of ontologies.
    OntologyProperty: URIRef  # The class of ontology properties.
    ReflexiveProperty: URIRef  # The class of reflexive properties.
    Restriction: URIRef  # The class of property restrictions.
    SymmetricProperty: URIRef  # The class of symmetric properties.
    TransitiveProperty: URIRef  # The class of transitive properties.

    # http://www.w3.org/2002/07/owl#AnnotationProperty
    backwardCompatibleWith: URIRef  # The annotation property that indicates that a given ontology is backward compatible with another ontology.
    deprecated: URIRef  # The annotation property that indicates that a given entity has been deprecated.
    incompatibleWith: URIRef  # The annotation property that indicates that a given ontology is incompatible with another ontology.
    priorVersion: URIRef  # The annotation property that indicates the predecessor ontology of a given ontology.
    versionInfo: URIRef  # The annotation property that provides version information for an ontology or another OWL construct.

    # http://www.w3.org/2002/07/owl#Class
    Nothing: URIRef  # This is the empty class.
    Thing: URIRef  # The class of OWL individuals.

    # http://www.w3.org/2002/07/owl#DatatypeProperty
    bottomDataProperty: URIRef  # The data property that does not relate any individual to any data value.
    topDataProperty: (
        URIRef  # The data property that relates every individual to every data value.
    )

    # http://www.w3.org/2002/07/owl#ObjectProperty
    bottomObjectProperty: (
        URIRef  # The object property that does not relate any two individuals.
    )
    topObjectProperty: URIRef  # The object property that relates every two individuals.

    # http://www.w3.org/2002/07/owl#OntologyProperty
    imports: URIRef  # The property that is used for importing other ontologies into a given ontology.
    versionIRI: URIRef  # The property that identifies the version IRI of an ontology.

    # http://www.w3.org/2000/01/rdf-schema#Datatype
    # NOTE: the following two elements don't appear in the OWL RDF documents but are defined in the OWL2 Recommentation
    # at https://www.w3.org/TR/owl2-syntax/#Datatype_Maps
    rational: URIRef  # The value space is the set of all rational numbers. The lexical form is numerator '/' denominator, where both are integers.
    real: URIRef  # The value space is the set of all real numbers. Does not directly provide any lexical forms.

    _NS = Namespace("http://www.w3.org/2002/07/owl#")



from __future__ import annotations

import typing as t
import urllib.parse


def sanitize_filename(filename: str) -> str | None:
    filename = filename.strip().strip("\"'")
    if not filename:
        return None
    filename = filename.replace("\\", "/").split("/")[-1]
    if not filename or filename in {".", ".."}:
        return None
    return filename


def extract_filename_from_content_disposition(header: str) -> str | None:
    """Extract filename from Content-Disposition header.

    Supports both ``filename=`` and RFC 5987 ``filename*=``; prefers ``filename*=``
    when present. Returned filenames are sanitized to prevent path traversal.
    """
    if not header:
        return None

    filename_star: str | None = None
    filename_plain: str | None = None

    for part in header.split(";"):
        part = part.strip()
        lower = part.lower()
        if lower.startswith("filename*="):
            value = part[len("filename*=") :].strip()
            value = value.strip().strip("\"'")
            try:
                charset, rest = value.split("'", 1)
                _language, encoded = rest.split("'", 1)
            except ValueError:
                charset = "utf-8"
                encoded = value
            try:
                raw = urllib.parse.unquote_to_bytes(encoded)
                decoded = raw.decode(charset, errors="replace")
            except LookupError:
                raw = urllib.parse.unquote_to_bytes(encoded)
                decoded = raw.decode("utf-8", errors="replace")
            filename_star = sanitize_filename(decoded)
        elif lower.startswith("filename="):
            value = part[len("filename=") :].strip()
            filename_plain = sanitize_filename(value)

    return filename_star or filename_plain


def infer_filename_from_fileobj(file_obj: t.Any, fallback: str) -> str:
    """Infer a safe filename from a file-like object's ``.name`` attribute."""
    name = getattr(file_obj, "name", None)
    if isinstance(name, str) and name:
        inferred = sanitize_filename(name)
        if inferred:
            return inferred
    return fallback
